from distutils import log as logger
from typing import Dict, List

from airflow.hooks.base import BaseHook
from sqlalchemy import FLOAT, and_, cast, column, func, select, text
from sqlalchemy.sql.expression import table as sqlatable

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table


class Check:
    def __init__(
        self,
        name: str,
        expression: str,
        threshold: float = 0,
    ) -> None:
        self.name = name
        self.expression = expression
        self.threshold = threshold

    def get_expression(self):
        return text(f"CASE WHEN {self.expression} THEN 0 ELSE 1 END AS {self.name}")

    def get_result(self):
        return cast(func.sum(column(self.name)), FLOAT) / func.count().label(
            self.name + "_result"
        )


class AgnosticBooleanCheck(SqlDecoratoratedOperator):
    template_fields = ("table",)

    def __init__(
        self,
        checks: List[Check],
        table: Table,
        max_rows_returned: int,
        **kwargs,
    ):
        """
        :param table: table to check
        :type table: Table
        :param checks: check class object, which represent boolean expression
        :type checks: Check
        :param max_rows_returned: number of row returned if the check fails.
        :type max_rows_returned: int
        """

        self.table = table
        self.max_rows_returned = max_rows_returned
        self.conn_id = table.conn_id
        self.checks = checks
        self.database = table.database

        task_id = table.table_name + "_" + "boolean_check"

        def handler_func(results):
            return results.fetchall()

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            conn_id=table.conn_id,
            database=table.database,
            schema=table.schema,
            warehouse=table.warehouse,
            task_id=task_id,
            op_args=(),
            python_callable=null_function,
            handler=handler_func,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn = BaseHook.get_connection(self.conn_id)
        self.conn_type = conn.conn_type  # type: ignore
        self.parameters = {"table": self.table}
        self.sql = self.prep_boolean_checks_query(self.table, self.checks, context)

        results = super().execute(context)
        failed_checks_names, failed_checks_index = self.get_failed_checks(results)
        if len(failed_checks_index) > 0:
            self.parameters = {"table": self.table, "limit": self.max_rows_returned}
            self.sql = self.prep_results(failed_checks_index)
            failed_rows = super().execute(context)
            logger.error("Failed rows %s", failed_rows)
            raise ValueError(
                "Some of the check(s) have failed %s", ",".join(failed_checks_names)
            )

        return self.table

    def get_failed_checks(self, results):
        failed_check_name = []
        failed_check_index = []

        for index in range(len(self.checks)):
            if self.checks[index].threshold < results[0][index]:
                failed_check_name.append(self.checks[index].name)
                failed_check_index.append(index)
        return failed_check_name, failed_check_index

    @staticmethod
    def get_expression(expression, name):
        return text(f"CASE WHEN {expression} THEN 0 ELSE 1 END AS {name}")

    def prep_boolean_checks_query(
        self, table: Table, checks: List[Check], context: Dict
    ):

        sqla_checks_object = []
        context = self._add_templates_to_context(context)
        for check in checks:
            prepared_exp = self.render_template(check.expression, context)
            sqla_checks_object.append(
                AgnosticBooleanCheck.get_expression(prepared_exp, check.name)
            )

        temp_table = (
            select(sqla_checks_object)
            .select_from(text(table.qualified_name()))
            .alias("check_table")
        )
        return select([check.get_result() for check in checks]).select_from(temp_table)

    def prep_results(self, results):
        return (
            select(["*"])
            .select_from(text("{{table}}"))
            .where(and_(*[text(self.checks[index].expression) for index in results]))
            .limit("{{limit}}")
        )


def boolean_check(table: Table, checks: List[Check] = [], max_rows_returned: int = 100):
    """
    :param table: table name
    :type table: str
    :param checks: check class object, which represent boolean expression
    :type checks: Check
    :param max_rows_returned: number of row returned if the check fails.
    :type max_rows_returned: int
    """

    return AgnosticBooleanCheck(
        table=table, checks=checks, max_rows_returned=max_rows_returned
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from astro import constants


class MissingPackage(object):
    def __init__(self, module_name, related_extras):
        self.module_name = module_name
        self.related_extras = related_extras

    def __getattr__(self, item):
        raise RuntimeError(
            f"Error loading the module {self.module_name},"
            f" please make sure all the dependencies are installed."
            f" try - pip install {constants.PYPI_PROJECT_NAME}[{self.related_extras}]"
        )


try:
    from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
except ModuleNotFoundError:
    BigQueryHook = MissingPackage(
        "airflow.providers.google.cloud.hooks.bigquery", "google"
    )

try:
    from airflow.providers.postgres.hooks.postgres import PostgresHook
except ModuleNotFoundError:
    PostgresHook = MissingPackage(
        "airflow.providers.postgres.hooks.postgres", "postgres"
    )

try:
    from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
except ModuleNotFoundError:
    SnowflakeHook = MissingPackage(
        "airflow.providers.snowflake.hooks.snowflake", "snowflake"
    )


try:
    from snowflake.connector import pandas_tools
except ModuleNotFoundError:
    pandas_tools = MissingPackage("snowflake-connector-python[pandas]", "postgres")


try:
    from boto3 import Session as BotoSession
except ModuleNotFoundError:
    BotoSession = MissingPackage("s3fs", "amazon")


try:
    from google.cloud.storage import Client as GCSClient
except ModuleNotFoundError:
    GCSClient = MissingPackage("apache-airflow-providers-google", "google")


try:
    from psycopg2 import sql as postgres_sql
except ModuleNotFoundError:
    postgres_sql = MissingPackage("psycopg2", "postgres")



from typing import Dict, List, Set

from airflow.hooks.base import BaseHook
from sqlalchemy import MetaData, case, func, or_, select
from sqlalchemy.sql.schema import Table as SqlaTable

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table
from astro.utils.schema_util import (
    get_error_string_for_multiple_dbs,
    get_table_name,
    tables_from_same_db,
)


class OutlierCheck:
    def __init__(
        self,
        name: str,
        columns_map: map,
        accepted_std_div: int,
        threshold: float,
    ) -> None:
        self.name = name
        self.columns_map = columns_map
        self.accepted_std_div = accepted_std_div
        self.threshold = threshold


class ChecksHandler:
    def __init__(self, checks: List[OutlierCheck], conn_type: str):
        self.checks = checks
        self.conn_type = conn_type

    def prepare_main_stats_sql(self, main_table: Table, main_table_sqla):
        select_expressions = []
        for check in self.checks:
            for key in check.columns_map:
                select_expressions.extend(
                    [
                        func.stddev(main_table_sqla.c.get(key)).label(
                            f"{check.name}_{key}_stddev"
                        ),
                        func.avg(main_table_sqla.c.get(key)).label(
                            f"{check.name}_{key}_avg"
                        ),
                    ]
                )
        return select(select_expressions).alias("main_stats_sql")

    def prepare_column_sql(self, check, main_stats, compare_table):
        column_sql = []
        for column in check.columns_map:
            main_table_col = column
            compare_table_col = compare_table.c.get(column)
            main_stats_check_avg = main_stats.c.get(
                f"{check.name}_{main_table_col}_avg"
            )
            main_stats_check_stddev = main_stats.c.get(
                f"{check.name}_{main_table_col}_stddev"
            )
            accepted_std_div = check.accepted_std_div

            statement = compare_table_col > (
                main_stats_check_avg + (main_stats_check_stddev * accepted_std_div)
            )
            column_sql.append(statement)
            statement = compare_table_col < (
                main_stats_check_avg - (main_stats_check_stddev * accepted_std_div)
            )
            column_sql.append(statement)

        return or_(*column_sql)

    def prepare_cases_sql(self, main_stats, compare_table_sqla):
        cases = []
        for check in self.checks:

            cases.append(
                case(
                    [
                        (
                            self.prepare_column_sql(
                                check, main_stats, compare_table_sqla
                            ),
                            1,
                        ),
                    ],
                    else_=0,
                ).label(check.name)
            )
        return cases

    def prepare_checks_sql(self, temp_table):
        aggregations = []
        for check in self.checks:
            aggregations.append(func.sum(temp_table.c.get(check.name)))
        return aggregations

    def prepare_comparison_sql(
        self, main_table: Table, compare_table: Table, engine, metadata_obj
    ):
        main_table_sqla = SqlaTable(
            get_table_name(main_table), metadata_obj, autoload_with=engine
        )
        compare_table_sqla = SqlaTable(
            get_table_name(compare_table), metadata_obj, autoload_with=engine
        )

        main_table_stats_sql = self.prepare_main_stats_sql(main_table, main_table_sqla)
        cases_sql = self.prepare_cases_sql(main_table_stats_sql, compare_table_sqla)
        cases_sql.append(compare_table_sqla)
        temp_table = select(cases_sql).alias("temp_table")
        checks_sql = self.prepare_checks_sql(temp_table)

        checks_sql.insert(0, func.count().label("total"))
        comparison_sql = select(checks_sql).select_from(temp_table)

        return comparison_sql

    def check_results(self, row, check, index):
        return row[index + 1]

    def evaluate_results(self, rows, max_rows_returned, conn_type):
        total_rows = rows[0][0]
        failed_rows_count = {
            check.name: self.check_results(rows[0], check, index)
            for index, check in enumerate(self.checks)
        }
        failed_checks = set()
        for check in self.checks:
            if failed_rows_count[check.name] / total_rows > check.threshold:
                failed_checks.add(check.name)
        return failed_checks

    def prepare_failed_checks_results(
        self,
        conn_type,
        main_table: Table,
        compare_table: Table,
        failed_checks: Set[str],
        max_rows_returned,
        engine,
        metadata_obj,
    ):
        main_table_sqla = SqlaTable(
            get_table_name(main_table), metadata_obj, autoload_with=engine
        )
        compare_table_sqla = SqlaTable(
            get_table_name(compare_table), metadata_obj, autoload_with=engine
        )

        main_stats = self.prepare_main_stats_sql(main_table, main_table_sqla)

        failed_checks_sql: List[tuple] = []
        for check in self.checks:
            if check.name in failed_checks:
                statement = (
                    select([compare_table_sqla])
                    .select_from(main_stats)
                    .select_from(compare_table_sqla)
                    .where(
                        self.prepare_column_sql(check, main_stats, compare_table_sqla)
                    )
                    .limit(max_rows_returned)
                )

                failed_checks_sql.append((check.name, statement, {}))
        return failed_checks_sql


class AgnosticStatsCheck(SqlDecoratoratedOperator):
    template_fields = ("table",)

    def __init__(
        self,
        checks: List[OutlierCheck],
        main_table: Table,
        compare_table: Table,
        max_rows_returned: int,
        **kwargs,
    ):
        """
        Operator to run Statistical checks on tables.

        :param main_table: main table
        :type main_table: table object
        :param compare_table: table to be compared
        :type compare_table: table object
        :param checks: check class object, which represent boolean expression
        :type checks: Check
        :param max_rows_returned: number of row returned if the check fails.
        :type max_rows_returned: int
        """

        self.main_table = main_table
        self.compare_table = compare_table
        self.max_rows_returned = max_rows_returned
        self.conn_id = main_table.conn_id
        self.checks = checks
        self.database = main_table.database

        task_id = main_table.table_name + "_" + "stats_check"

        if not tables_from_same_db([main_table, compare_table]):
            raise ValueError(
                get_error_string_for_multiple_dbs([main_table, compare_table])
            )

        def null_function():
            pass

        def handler_func(results):
            return results.fetchall()

        super().__init__(
            raw_sql=True,
            parameters={},
            conn_id=main_table.conn_id,
            database=main_table.database,
            schema=main_table.schema,
            warehouse=main_table.warehouse,
            task_id=task_id,
            op_args=(),
            handler=handler_func,
            python_callable=null_function,
            **kwargs,
        )

    def execute(self, context: Dict):
        conn_type = BaseHook.get_connection(self.conn_id).conn_type  # type: ignore
        checkHandler = ChecksHandler(self.checks, conn_type)
        metadata = MetaData()
        self.sql = checkHandler.prepare_comparison_sql(
            main_table=self.main_table,
            compare_table=self.compare_table,
            engine=self.get_sql_alchemy_engine(),
            metadata_obj=metadata,
        )

        results = super().execute(context)
        failed_checks = checkHandler.evaluate_results(
            results, self.max_rows_returned, conn_type
        )
        if len(failed_checks) > 0:
            failed_check_sql = checkHandler.prepare_failed_checks_results(
                conn_type=conn_type,
                main_table=self.main_table,
                compare_table=self.compare_table,
                failed_checks=failed_checks,
                max_rows_returned=self.max_rows_returned,
                engine=self.get_sql_alchemy_engine(),
                metadata_obj=metadata,
            )
            failed_values = {}
            for check_query in failed_check_sql:
                check, self.sql, self.parameters = check_query
                failed_values[check] = super().execute(context)

            raise ValueError("Stats Check Failed. %s", failed_values)


def stats_check(
    main_table: Table,
    compare_table: Table,
    checks: List[OutlierCheck] = [],
    max_rows_returned: int = 100,
):
    """
    :param main_table: main table
    :type main_table: table object
    :param compare_table: table to be compared
    :type compare_table: table object
    :param checks: check class object, which represent boolean expression
    :type checks: Check
    :param max_rows_returned: number of row returned if the check fails.
    :type max_rows_returned: int
    """

    return AgnosticStatsCheck(
        main_table=main_table,
        compare_table=compare_table,
        checks=checks,
        max_rows_returned=max_rows_returned,
    )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import inspect
from typing import Optional

import pandas

from astro.sql.table import Table
from astro.utils.schema_util import get_schema


class TableHandler:
    def _set_variables_from_first_table(self):
        """
        When we create our SQL operation, we run with the assumption that the first table given is the "main table".
        This means that a user doesn't need to define default conn_id, database, etc. in the function unless they want
        to create default values.
        """
        first_table: Optional[Table] = None
        if self.op_args:
            table_index = [x for x, t in enumerate(self.op_args) if type(t) == Table]
            if table_index:
                first_table = self.op_args[table_index[0]]
        elif not first_table:
            table_kwargs = [
                x
                for x in inspect.signature(self.python_callable).parameters.values()
                if (
                    x.annotation == Table
                    and type(self.op_kwargs[x.name]) == Table
                    or x.annotation == pandas.DataFrame
                    and type(self.op_kwargs[x.name]) == Table
                )
            ]
            if table_kwargs:
                first_table = self.op_kwargs[table_kwargs[0].name]

        # If there is no first table via op_ags or kwargs, we check the parameters
        elif not first_table:
            if self.parameters:
                param_tables = [t for t in self.parameters.values() if type(t) == Table]
                if param_tables:
                    first_table = param_tables[0]

        if first_table:
            self.conn_id = first_table.conn_id or self.conn_id
            self.database = first_table.database or self.database
            self.schema = first_table.schema or self.schema
            self.warehouse = first_table.warehouse or self.warehouse
            self.role = first_table.role or self.role

    def populate_output_table(self):
        self.output_table.conn_id = self.output_table.conn_id or self.conn_id
        self.output_table.database = self.output_table.database or self.database
        self.output_table.warehouse = self.output_table.warehouse or self.warehouse
        self.output_table.schema = self.output_table.schema or get_schema()



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from astro.sql.table import Table
from astro.utils.dependencies import PostgresHook


def add_templates_to_context(parameters, context):
    for k, v in parameters.items():
        if type(v) == Table:
            final_name = v.schema + "." + v.table_name if v.schema else v.table_name
            context[k] = final_name
        else:
            context[k] = ":" + k
    return context


def create_sql_engine(postgres_conn_id, database):
    hook = PostgresHook(postgres_conn_id=postgres_conn_id, schema=database)
    engine = hook.get_sqlalchemy_engine()
    engine.url.database = database
    return engine



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import importlib
from typing import Dict, List

from sqlalchemy import MetaData, cast, column, insert, select
from sqlalchemy.sql.schema import Table as SqlaTable

from astro.sql.operators.sql_decorator import SqlDecoratoratedOperator
from astro.sql.table import Table
from astro.utils.schema_util import (
    get_error_string_for_multiple_dbs,
    get_table_name,
    tables_from_same_db,
)
from astro.utils.table_handler import TableHandler
from astro.utils.task_id_helper import get_unique_task_id


class SqlAppendOperator(SqlDecoratoratedOperator, TableHandler):
    template_fields = ("main_table", "append_table")

    def __init__(
        self,
        append_table: Table,
        main_table: Table,
        columns: List[str] = [],
        casted_columns: dict = {},
        **kwargs,
    ):
        self.append_table = append_table
        self.main_table = main_table
        self.sql = ""

        self.columns = columns
        self.casted_columns = casted_columns
        task_id = get_unique_task_id("append_table")

        def null_function():
            pass

        super().__init__(
            raw_sql=True,
            parameters={},
            task_id=kwargs.get("task_id") or task_id,
            op_args=(),
            python_callable=null_function,
            handler=lambda x: None,
            **kwargs,
        )

    def execute(self, context: Dict):
        if not tables_from_same_db([self.append_table, self.main_table]):
            raise ValueError(
                get_error_string_for_multiple_dbs([self.append_table, self.main_table])
            )

        self.main_table.conn_id = self.main_table.conn_id or self.append_table.conn_id
        self.conn_id = self.main_table.conn_id or self.append_table.conn_id
        self.database = self.main_table.database or self.append_table.database
        self.warehouse = self.main_table.warehouse or self.append_table.warehouse
        self.schema = self.main_table.schema or self.append_table.schema
        self.sql = self.append(
            main_table=self.main_table,
            append_table=self.append_table,
            columns=self.columns,
            casted_columns=self.casted_columns,
            conn_id=self.main_table.conn_id,
        )
        return super().execute(context)

    def append(
        self, main_table: Table, columns, casted_columns, append_table: Table, conn_id
    ):
        engine = self.get_sql_alchemy_engine()
        metadata = MetaData()
        # TO Do - fix bigquery and postgres reflection table issue.
        main_table_sqla = SqlaTable(
            get_table_name(main_table), metadata, autoload_with=engine
        )
        append_table_sqla = SqlaTable(
            get_table_name(append_table), metadata, autoload_with=engine
        )

        column_names = [column(c) for c in columns]
        sqlalchemy = importlib.import_module("sqlalchemy")
        casted_fields = [
            cast(column(k), getattr(sqlalchemy, v)) for k, v in casted_columns.items()
        ]
        main_columns = [k for k, v in casted_columns.items()]
        main_columns.extend([c for c in columns])

        if len(column_names) + len(casted_fields) == 0:
            column_names = [column(c) for c in append_table_sqla.c.keys()]
            main_columns = column_names

        column_names.extend(casted_fields)
        sel = select(column_names).select_from(append_table_sqla)
        return insert(main_table_sqla).from_select(main_columns, sel)



import os
from typing import List

from airflow.hooks.base import BaseHook

from astro.sql.table import Table
from astro.utils.dependencies import postgres_sql


def schema_exists(hook, schema, conn_type):
    if conn_type in ["postgresql", "postgres"]:
        created_schemas = hook.run(
            "SELECT schema_name FROM information_schema.schemata;",
            handler=lambda x: [y[0] for y in x.fetchall()],
        )
        return schema.upper() in [c.upper() for c in created_schemas]
    elif conn_type == "snowflake":
        created_schemas = [
            x["SCHEMA_NAME"]
            for x in hook.run("SELECT SCHEMA_NAME from information_schema.schemata;")
        ]
        return schema.upper() in [c.upper() for c in created_schemas]
    return False


def create_schema_query(conn_type, hook, schema_id, user):

    if conn_type in ["postgresql", "postgres"]:
        return (
            postgres_sql.SQL(
                "CREATE SCHEMA IF NOT EXISTS {schema} AUTHORIZATION {user}"
            )
            .format(
                schema=postgres_sql.Identifier(schema_id),
                user=postgres_sql.Identifier(user),
            )
            .as_string(hook.get_conn())
        )
    elif conn_type in ["snowflake", "google_cloud_platform", "bigquery", "sqlite"]:
        return f"CREATE SCHEMA IF NOT EXISTS {schema_id}"


def get_schema():
    return os.getenv("AIRFLOW__ASTRO__SQL_SCHEMA") or "tmp_astro"


def get_table_name(table: Table):
    conn_type = BaseHook.get_connection(table.conn_id).conn_type
    if conn_type in ["bigquery"]:
        return table.qualified_name()
    return table.table_name


def tables_from_same_db(tables: List[Table]):
    """
    Validate that the tables belong to same db by checking connection type.
    :param tables: List of table
    :return: Boolean
    """
    conn_ids = set()
    for table in tables:
        if table.conn_id:
            conn_ids.add(table.conn_id)
    return len(conn_ids) == 1


def get_error_string_for_multiple_dbs(tables: List[Table]):
    """
    Get error string for tables belonging to multiple databases.
    :param tables: list of table
    :return: String: error string
    """
    return f'Tables should belong to same db {", ".join([f"{table.table_name}: {table.conn_id}" for table in tables])}'



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
from typing import Optional, Union

from airflow.hooks.sqlite_hook import SqliteHook
from pandas import DataFrame
from pandas.io.sql import SQLDatabase

from astro.utils.dependencies import (
    BigQueryHook,
    PostgresHook,
    SnowflakeHook,
    pandas_tools,
)
from astro.utils.schema_util import create_schema_query, schema_exists


def move_dataframe_to_sql(
    output_table_name,
    conn_id,
    database,
    schema,
    warehouse,
    conn_type,
    df: DataFrame,
    user,
    chunksize,
):
    hook_kwargs = {
        "postgresql": {"postgres_conn_id": conn_id, "schema": database},
        "postgres": {"postgres_conn_id": conn_id, "schema": database},
        "snowflake": {
            "snowflake_conn_id": conn_id,
            "database": database,
            "schema": schema,
            "warehouse": warehouse,
        },
        "bigquery": {"use_legacy_sql": False, "gcp_conn_id": conn_id},
        "sqlite": {"sqlite_conn_id": conn_id},
    }
    try:
        hook_class = {
            "postgresql": PostgresHook,
            "postgres": PostgresHook,
            "snowflake": SnowflakeHook,
            "bigquery": BigQueryHook,
            "sqlite": SqliteHook,
        }[conn_type]
    except KeyError:
        raise ValueError(
            f"The conn_id {conn_id} is of unsupported type {conn_type}. Current supported types: {list(hook_class.keys())}"
        )
    else:
        hook = hook_class(**hook_kwargs[conn_type])

    if database:
        hook.database = database

    if conn_type != "sqlite" and not schema_exists(
        hook=hook, schema=schema, conn_type=conn_type
    ):
        schema_query = create_schema_query(
            conn_type=conn_type, hook=hook, schema_id=schema, user=user
        )
        hook.run(schema_query)

    if conn_type == "snowflake":

        db = SQLDatabase(engine=hook.get_sqlalchemy_engine())
        # make columns uppercase to prevent weird errors in snowflake
        df.columns = df.columns.str.upper()
        db.prep_table(
            df,
            output_table_name.lower(),
            schema=schema,
            if_exists="replace",
            index=False,
        )
        pandas_tools.write_pandas(
            hook.get_conn(),
            df,
            output_table_name,
            chunk_size=chunksize,
            quote_identifiers=False,
        )
    elif conn_type == "bigquery":
        df.to_gbq(
            f"{schema}.{output_table_name}",
            if_exists="replace",
            chunksize=chunksize,
            project_id=hook.project_id,
        )
    elif conn_type == "sqlite":
        df.to_sql(
            output_table_name,
            con=hook.get_sqlalchemy_engine(),
            if_exists="replace",
            chunksize=chunksize,
            method="multi",
            index=False,
        )
    else:
        df.to_sql(
            output_table_name,
            con=hook.get_sqlalchemy_engine(),
            schema=schema,
            if_exists="replace",
            chunksize=chunksize,
            method="multi",
            index=False,
        )



"""
Copyright Astronomer, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from airflow.hooks.base import BaseHook
from airflow.models import DagRun, TaskInstance


class Table:
    template_fields = (
        "table_name",
        "conn_id",
        "database",
        "schema",
        "warehouse",
        "role",
    )

    def __init__(
        self,
        table_name="",
        conn_id=None,
        database=None,
        schema=None,
        warehouse=None,
        role=None,
    ):
        self.table_name = table_name
        self.conn_id = conn_id
        self.database = database
        self.schema = schema
        self.warehouse = warehouse
        self.role = role

    def identifier_args(self):
        return (self.schema, self.table_name) if self.schema else (self.table_name,)

    def qualified_name(self):
        conn = BaseHook.get_connection(self.conn_id)
        self.conn_type = conn.conn_type  # type: ignore
        if self.conn_type == "sqlite":
            return self.table_name
        else:
            return (
                self.schema + "." + self.table_name if self.schema else self.table_name
            )

    def __str__(self):
        return f"Table(table_name={self.table_name}, database={self.database}, schema={self.schema}, conn_id={self.conn_id}, warehouse={self.warehouse})"


class TempTable(Table):
    def __init__(self, conn_id=None, database=None, warehouse=""):
        super().__init__(
            table_name="", conn_id=conn_id, database=database, warehouse=warehouse
        )

    def to_table(self, table_name: str, schema: str) -> Table:
        return Table(
            table_name=table_name,
            conn_id=self.conn_id,
            database=self.database,
            warehouse=self.warehouse,
            schema=schema,
        )


def create_table_name(context) -> str:
    ti: TaskInstance = context["ti"]
    dag_run: DagRun = ti.get_dagrun()
    table_name = f"{dag_run.dag_id}_{ti.task_id}_{dag_run.id}".replace(
        "-", "_"
    ).replace(".", "__")
    if not table_name.isidentifier():
        table_name = f'"{table_name}"'
    return table_name
